@@ -18,7 +18,8 @@
 
 extern obs_websocket_vendor vendor;
 
-DownstreamKeyer::DownstreamKeyer(int channel, QString name, obs_view_t *v, get_transitions_callback_t gt, void *gtd)
+DownstreamKeyer::DownstreamKeyer(int channel, QString name, obs_view_t *v, obs_canvas_t *c, get_transitions_callback_t gt,
+				 void *gtd)
 	: outputChannel(channel),
 	  transition(nullptr),
 	  showTransition(nullptr),
@@ -28,6 +29,7 @@ DownstreamKeyer::DownstreamKeyer(int channel, QString name, obs_view_t *v, get_t
 	  showTransitionDuration(300),
 	  hideTransitionDuration(300),
 	  view(v),
+	  canvas(c),
 	  get_transitions(gt),
 	  get_transitions_data(gtd)
 {
@@ -156,6 +158,8 @@ DownstreamKeyer::~DownstreamKeyer()
 {
 	if (view) {
 		//obs_view_set_source(view, outputChannel, nullptr);
+	} else if (canvas) {
+		//obs_canvas_set_channel(canvas, outputChannel, nullptr);
 	} else {
 		obs_set_output_source(outputChannel, nullptr);
 	}
@@ -212,6 +216,20 @@ void DownstreamKeyer::on_actionAddScene_triggered()
 		} else {
 			obs_source_release(source);
 		}
+	} else if (canvas) {
+		obs_source_t *source = obs_canvas_get_channel(canvas, 0);
+		if (source && obs_source_get_type(source) == OBS_SOURCE_TYPE_TRANSITION) {
+			obs_source_t *ts = obs_transition_get_active_source(source);
+			if (ts) {
+				obs_source_release(source);
+				source = ts;
+			}
+		}
+		if (source && obs_source_is_scene(source)) {
+			scene = source;
+		} else {
+			obs_source_release(source);
+		}
 	} else {
 		scene = obs_frontend_preview_program_mode_active() ? obs_frontend_get_current_preview_scene()
 								   : obs_frontend_get_current_scene();
@@ -220,7 +238,8 @@ void DownstreamKeyer::on_actionAddScene_triggered()
 		return;
 	auto sceneName = QT_UTF8(obs_source_get_name(scene));
 	if (scenesList->findItems(sceneName, Qt::MatchFixedString).count() == 0) {
-		add_scene(sceneName, scene);
+		const auto currentRow = scenesList->currentRow();
+		add_scene(sceneName, scene, currentRow);
 	}
 
 	obs_source_release(scene);
@@ -262,7 +281,9 @@ void DownstreamKeyer::apply_source(obs_source_t *const newSource)
 		hideTimer.setInterval(hideAfter);
 		hideTimer.start();
 	}
-	obs_source_t *prevSource = view ? obs_view_get_source(view, outputChannel) : obs_get_output_source(outputChannel);
+	obs_source_t *prevSource = view     ? obs_view_get_source(view, outputChannel)
+				   : canvas ? obs_canvas_get_channel(canvas, outputChannel)
+					    : obs_get_output_source(outputChannel);
 	obs_source_t *prevTransition = nullptr;
 	if (prevSource && obs_source_get_type(prevSource) == OBS_SOURCE_TYPE_TRANSITION) {
 		prevTransition = prevSource;
@@ -302,6 +323,8 @@ void DownstreamKeyer::apply_source(obs_source_t *const newSource)
 		if (!newTransition) {
 			if (view) {
 				obs_view_set_source(view, outputChannel, newSource);
+			} else if (canvas) {
+				obs_canvas_set_channel(canvas, outputChannel, newSource);
 			} else {
 				obs_set_output_source(outputChannel, newSource);
 			}
@@ -313,6 +336,8 @@ void DownstreamKeyer::apply_source(obs_source_t *const newSource)
 			if (prevTransition != newTransition) {
 				if (view) {
 					obs_view_set_source(view, outputChannel, newTransition);
+				} else if (canvas) {
+					obs_canvas_set_channel(canvas, outputChannel, newTransition);
 				} else {
 					obs_set_output_source(outputChannel, newTransition);
 				}
@@ -336,7 +361,9 @@ void DownstreamKeyer::apply_source(obs_source_t *const newSource)
 void DownstreamKeyer::apply_selected_source()
 {
 	const auto l = scenesList->selectedItems();
-	const auto newSource = l.count() ? obs_get_source_by_name(QT_TO_UTF8(l.value(0)->text())) : nullptr;
+	const auto newSource = l.count() ? (canvas ? obs_canvas_get_source_by_name(canvas, QT_TO_UTF8(l.value(0)->text()))
+						   : obs_get_source_by_name(QT_TO_UTF8(l.value(0)->text())))
+					 : nullptr;
 
 	apply_source(newSource);
 	obs_source_release(newSource);
@@ -462,13 +489,17 @@ void DownstreamKeyer::SetTransition(const char *transition_name, enum transition
 		overrideTransition = newTransition;
 	else
 		transition = newTransition;
-	obs_source_t *prevSource = view ? obs_view_get_source(view, outputChannel) : obs_get_output_source(outputChannel);
+	obs_source_t *prevSource = view     ? obs_view_get_source(view, outputChannel)
+				   : canvas ? obs_canvas_get_channel(canvas, outputChannel)
+					    : obs_get_output_source(outputChannel);
 	if (oldTransition && prevSource == oldTransition) {
 		if (newTransition) {
 			//swap transition
 			obs_transition_swap_begin(newTransition, oldTransition);
 			if (view) {
 				obs_view_set_source(view, outputChannel, newTransition);
+			} else if (canvas) {
+				obs_canvas_set_channel(canvas, outputChannel, newTransition);
 			} else {
 				obs_set_output_source(outputChannel, newTransition);
 			}
@@ -476,9 +507,12 @@ void DownstreamKeyer::SetTransition(const char *transition_name, enum transition
 		} else {
 			auto item = scenesList->currentItem();
 			if (item) {
-				auto scene = obs_get_source_by_name(QT_TO_UTF8(item->text()));
+				auto scene = canvas ? obs_canvas_get_source_by_name(canvas, QT_TO_UTF8(item->text()))
+						    : obs_get_source_by_name(QT_TO_UTF8(item->text()));
 				if (view) {
 					obs_view_set_source(view, outputChannel, scene);
+				} else if (canvas) {
+					obs_canvas_set_channel(canvas, outputChannel, scene);
 				} else {
 					obs_set_output_source(outputChannel, scene);
 				}
@@ -486,6 +520,8 @@ void DownstreamKeyer::SetTransition(const char *transition_name, enum transition
 			} else {
 				if (view) {
 					obs_view_set_source(view, outputChannel, nullptr);
+				} else if (canvas) {
+					obs_canvas_set_channel(canvas, outputChannel, nullptr);
 				} else {
 					obs_set_output_source(outputChannel, nullptr);
 				}
@@ -546,7 +582,9 @@ void DownstreamKeyer::SceneChanged(std::string scene)
 		apply_source(nullptr);
 		return;
 	} else {
-		obs_source_t *prevSource = view ? obs_view_get_source(view, outputChannel) : obs_get_output_source(outputChannel);
+		obs_source_t *prevSource = view     ? obs_view_get_source(view, outputChannel)
+					   : canvas ? obs_canvas_get_channel(canvas, outputChannel)
+						    : obs_get_output_source(outputChannel);
 		if (prevSource && obs_source_get_type(prevSource) == OBS_SOURCE_TYPE_TRANSITION) {
 			obs_source_t *prevTransition = prevSource;
 			prevSource = obs_transition_get_active_source(prevTransition);
@@ -585,11 +623,14 @@ void DownstreamKeyer::Load(obs_data_t *data)
 			const auto source_name = obs_data_get_string(sceneData, "name");
 			const auto item = new QListWidgetItem(QT_UTF8(source_name));
 			scenesList->addItem(item);
-			obs_source_t *source = obs_get_source_by_name(source_name);
+			obs_source_t *source = canvas ? obs_canvas_get_source_by_name(canvas, source_name)
+						      : obs_get_source_by_name(source_name);
 			if (item->text() == sceneName) {
 				if (source) {
 					if (view) {
 						obs_view_set_source(view, outputChannel, source);
+					} else if (canvas) {
+						obs_canvas_set_channel(canvas, outputChannel, source);
 					} else {
 						obs_set_output_source(outputChannel, source);
 					}
@@ -774,6 +815,20 @@ void DownstreamKeyer::AddExcludeScene(const char *scene_name)
 		} else {
 			obs_source_release(source);
 		}
+	} else if (canvas) {
+		obs_source_t *source = obs_canvas_get_channel(canvas, 0);
+		if (source && obs_source_get_type(source) == OBS_SOURCE_TYPE_TRANSITION) {
+			obs_source_t *ts = obs_transition_get_active_source(source);
+			if (ts) {
+				obs_source_release(source);
+				source = ts;
+			}
+		}
+		if (source && obs_source_is_scene(source)) {
+			scene = source;
+		} else {
+			obs_source_release(source);
+		}
 	} else {
 		scene = obs_frontend_get_current_scene();
 	}
@@ -801,6 +856,20 @@ void DownstreamKeyer::RemoveExcludeScene(const char *scene_name)
 		} else {
 			obs_source_release(source);
 		}
+	} else if (canvas) {
+		obs_source_t *source = obs_canvas_get_channel(canvas, 0);
+		if (source && obs_source_get_type(source) == OBS_SOURCE_TYPE_TRANSITION) {
+			obs_source_t *ts = obs_transition_get_active_source(source);
+			if (ts) {
+				obs_source_release(source);
+				source = ts;
+			}
+		}
+		if (source && obs_source_is_scene(source)) {
+			scene = source;
+		} else {
+			obs_source_release(source);
+		}
 	} else {
 		scene = obs_frontend_get_current_scene();
 	}
@@ -814,6 +883,19 @@ bool DownstreamKeyer::IsSceneExcluded(const char *scene_name)
 	return exclude_scenes.find(scene_name) != exclude_scenes.end();
 }
 
+QString DownstreamKeyer::GetScene()
+{
+	for (int i = 0; i < scenesList->count(); i++) {
+		const auto item = scenesList->item(i);
+		if (!item)
+			continue;
+		if (!item->isSelected())
+			continue;
+		return item->text();
+	}
+	return "";
+}
+
 bool DownstreamKeyer::SwitchToScene(QString scene_name)
 {
 	if (scene_name.isEmpty()) {
@@ -834,10 +916,14 @@ bool DownstreamKeyer::SwitchToScene(QString scene_name)
 	return false;
 }
 
-void DownstreamKeyer::add_scene(QString scene_name, obs_source_t *s)
+void DownstreamKeyer::add_scene(QString scene_name, obs_source_t *s, int insertBeforeRow)
 {
 	const auto item = new QListWidgetItem(scene_name);
-	scenesList->addItem(item);
+	int scenesListCount = scenesList->count();
+	if ((insertBeforeRow > scenesListCount) || (insertBeforeRow < 0)) {
+		insertBeforeRow = scenesListCount;
+	}
+	scenesList->insertItem(insertBeforeRow, item);
 
 	std::string enable_hotkey = obs_module_text("EnableDSK");
 	enable_hotkey += " ";
@@ -853,7 +939,7 @@ void DownstreamKeyer::add_scene(QString scene_name, obs_source_t *s)
 	}
 }
 
-bool DownstreamKeyer::AddScene(QString scene_name)
+bool DownstreamKeyer::AddScene(QString scene_name, int insertBeforeRow)
 {
 	if (scene_name.isEmpty()) {
 		return false;
@@ -863,10 +949,9 @@ bool DownstreamKeyer::AddScene(QString scene_name)
 	}
 	auto nameUtf8 = scene_name.toUtf8();
 	auto name = nameUtf8.constData();
-	auto s = obs_get_source_by_name(name);
+	auto s = canvas ? obs_canvas_get_source_by_name(canvas, name) : obs_get_source_by_name(name);
 	if (obs_source_is_scene(s)) {
-
-		add_scene(scene_name, s);
+		add_scene(scene_name, s, insertBeforeRow);
 		obs_source_release(s);
 		return true;
 	}
@@ -903,7 +988,9 @@ void DownstreamKeyer::SetOutputChannel(int oc)
 {
 	if (oc == outputChannel)
 		return;
-	obs_source_t *prevSource = view ? obs_view_get_source(view, outputChannel) : obs_get_output_source(outputChannel);
+	obs_source_t *prevSource = view     ? obs_view_get_source(view, outputChannel)
+				   : canvas ? obs_canvas_get_channel(canvas, outputChannel)
+					    : obs_get_output_source(outputChannel);
 	obs_source_t *prevTransition = nullptr;
 	if (prevSource && obs_source_get_type(prevSource) == OBS_SOURCE_TYPE_TRANSITION) {
 		prevTransition = prevSource;
@@ -914,6 +1001,8 @@ void DownstreamKeyer::SetOutputChannel(int oc)
 		    prevTransition == overrideTransition) {
 			if (view) {
 				obs_view_set_source(view, outputChannel, nullptr);
+			} else if (canvas) {
+				obs_canvas_set_channel(canvas, outputChannel, nullptr);
 			} else {
 				obs_set_output_source(outputChannel, nullptr);
 			}
@@ -923,10 +1012,14 @@ void DownstreamKeyer::SetOutputChannel(int oc)
 		}
 	} else if (prevSource) {
 		const auto l = scenesList->selectedItems();
-		const auto newSource = l.count() ? obs_get_source_by_name(QT_TO_UTF8(l.value(0)->text())) : nullptr;
+		const auto newSource = l.count() ? (canvas ? obs_canvas_get_source_by_name(canvas, QT_TO_UTF8(l.value(0)->text()))
+							   : obs_get_source_by_name(QT_TO_UTF8(l.value(0)->text())))
+						 : nullptr;
 		if (prevSource == newSource) {
 			if (view) {
 				obs_view_set_source(view, outputChannel, nullptr);
+			} else if (canvas) {
+				obs_canvas_set_channel(canvas, outputChannel, nullptr);
 			} else {
 				obs_set_output_source(outputChannel, nullptr);
 			}
@@ -940,6 +1033,8 @@ void DownstreamKeyer::SetOutputChannel(int oc)
 	if (prevTransition) {
 		if (view) {
 			obs_view_set_source(view, outputChannel, prevTransition);
+		} else if (canvas) {
+			obs_canvas_set_channel(canvas, outputChannel, prevTransition);
 		} else {
 			obs_set_output_source(outputChannel, prevTransition);
 		}
