@@ -47,6 +47,7 @@ private:
 	obs_hotkey_pair_id tie_hotkey_id;
 	std::set<std::string> exclude_scenes;
 	obs_view_t *view = nullptr;
+	obs_canvas_t *canvas = nullptr;
 	get_transitions_callback_t get_transitions = nullptr;
 	void *get_transitions_data = nullptr;
 
@@ -74,8 +75,8 @@ private slots:
 signals:
 
 public:
-	DownstreamKeyer(int channel, QString name, obs_view_t *view = nullptr, get_transitions_callback_t get_transitions = nullptr,
-			void *get_transitions_data = nullptr);
+	DownstreamKeyer(int channel, QString name, obs_view_t *view = nullptr, obs_canvas_t *canvas = nullptr,
+			get_transitions_callback_t get_transitions = nullptr, void *get_transitions_data = nullptr);
 	~DownstreamKeyer();
 
 	void Save(obs_data_t *data);
@@ -90,9 +91,10 @@ public:
 	void AddExcludeScene(const char *scene_name);
 	void RemoveExcludeScene(const char *scene_name);
 	bool IsSceneExcluded(const char *scene_name);
+	QString GetScene();
 	bool SwitchToScene(QString scene_name);
-	void add_scene(QString scene_name, obs_source_t *s);
-	bool AddScene(QString scene_name);
+	void add_scene(QString scene_name, obs_source_t *s, int insertBeforeRow);
+	bool AddScene(QString scene_name, int insertBeforeRow);
 	bool RemoveScene(QString scene_name);
 	void SetTie(bool tie);
 	void SetOutputChannel(int outputChannel);
